
// SDK includes
#include <bluetooth/gatt.h>
#include <bluetooth/gatt_dm.h>
#include <bluetooth/scan.h>
#include <bluetooth/uuid.h>

#include <logging/log.h>
#include <misc/byteorder.h>
#include <stdio.h>
#include <zephyr.h>

// Project includes
#include "ble.h"

static bool do_discover = false;

LOG_MODULE_REGISTER(ble, LOG_LEVEL_DBG);

static struct bt_conn *m_conn;

bool adv_report_get_name(const struct net_buf_simple *ad, char name[32]) {

    uint16_t index;

    // extract name if present
    index = 0;
    while (index < ad->len) {
        uint8_t field_len  = ad->data[index];
        uint8_t field_type = ad->data[index + 1];

        if (field_type == BT_DATA_NAME_COMPLETE ||
            field_type == BT_DATA_NAME_SHORTENED) {
            memcpy(name, &ad->data[index + 2], field_len - 1);
            name[field_len - 1] = '\0';
            return true;
        }
        index += field_len + 1;
    }
    return false;
}
static u8_t on_notification_received(struct bt_conn *conn,
                                     struct bt_gatt_subscribe_params *params,
                                     const void *data, u16_t length) {

    LOG_HEXDUMP_INF(data, length, "notification: ");

    return BT_GATT_ITER_CONTINUE;
}

static void discovery_service_not_found(struct bt_conn *conn, void *ctx) {
    LOG_ERR("discovery_service_not_found");

    int err = bt_conn_disconnect(conn, BT_HCI_ERR_REMOTE_USER_TERM_CONN);
    if (err != 0) {
        LOG_ERR("failed to ble disconn: %d", err);
    }
}

static void discovery_error_found(struct bt_conn *conn, int err, void *ctx) {
    LOG_ERR("discovery_error_found %d", err);

    err = bt_conn_disconnect(conn, BT_HCI_ERR_REMOTE_USER_TERM_CONN);
    if (err != 0) {
        LOG_ERR("failed to ble disconn: %d", err);
    }
}

static void discovery_completed(struct bt_gatt_dm *disc, void *ctx) {

    LOG_INF("discovery completed, subscribing");

    static struct bt_gatt_subscribe_params param = {
        .notify       = on_notification_received,
        .value        = BT_GATT_CCC_NOTIFY,
        .ccc_handle   = 0x0011,
        .value_handle = 0x0010};

    LOG_INF("subscribing without service discovery.");
    LOG_INF("cccd handle 0x%04x, val handle 0x%04x", param.ccc_handle, param.value_handle);
    int err = bt_gatt_subscribe(m_conn, &param);
    if (err) {
        LOG_ERR("Subscribe failed (err %d)", err);
    }
}

static struct bt_gatt_dm_cb discovery_cb = {
    .completed         = discovery_completed,
    .service_not_found = discovery_service_not_found,
    .error_found       = discovery_error_found,
};

static void connected(struct bt_conn *conn, u8_t conn_err) {
    m_conn = conn;

    char addr[BT_ADDR_LE_STR_LEN];
    if (conn_err) {
        LOG_ERR("Failed to connect to %s (%u)", addr, conn_err);
        return;
    }

    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
    LOG_INF("Connected to: %s", addr);

    if (do_discover) {
        LOG_INF("starting service discovery...");
        int err = bt_gatt_dm_start(conn, NULL, &discovery_cb, NULL);
        if (err) {
            LOG_ERR("Could not start service discovery, err %d", err);
        }
    } else {
        // hard-coded cccd and value handle for ble_app_hrs, sdk 15.2, pca10040
        static struct bt_gatt_subscribe_params param = {
            .notify       = on_notification_received,
            .value        = BT_GATT_CCC_NOTIFY,
            .ccc_handle   = 0x0011,
            .value_handle = 0x0010};

        LOG_INF("subscribing without service discovery.");
        LOG_INF("cccd handle 0x%04x, val handle 0x%04x", param.ccc_handle, param.value_handle);
        int err = bt_gatt_subscribe(conn, &param);
        if (err) {
            LOG_ERR("Subscribe failed (err %d)", err);
        }
    }
}

static void disconnected(struct bt_conn *conn, u8_t reason) {

    char addr[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

    LOG_INF("Disconnected from: %s", addr);
}

void scan_connecting_error(struct bt_scan_device_info *device_info) {
    LOG_ERR("Connection to peer failed!");
}

static void device_found(const bt_addr_le_t *addr, s8_t rssi, u8_t type,
                         struct net_buf_simple *ad) {

    // ignore if not connectable
    if (type != BT_LE_ADV_IND && type != BT_LE_ADV_DIRECT_IND) {
        return;
    }

    // ignore if does not advertise with a name
    char name[128];
    bool has_name = adv_report_get_name(ad, name);
    if (!has_name) {
        return;
    }

    char addr_str[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(addr, addr_str, sizeof(addr_str));
    LOG_INF("Device found: %s (RSSI %d) - %s", addr_str, rssi, name);
    
    // check for hard-coded peripheral name (debug)
    if (strcmp("Nordic_HRM", name) != 0) {
        return;
    }

    int err = bt_le_scan_stop();
    if (err != 0) {
        LOG_ERR("failed to bt_le_scan_stop: err %d\"n", err);
        return;
    }

    LOG_INF("connecting...");
    m_conn = bt_conn_create_le(addr, BT_LE_CONN_PARAM_DEFAULT);
    if (m_conn == NULL) {
        LOG_ERR("failed to create LE connection, errno %d", errno);
    }
}

BT_SCAN_CB_INIT(scan_cb, NULL, NULL, scan_connecting_error, NULL);

static void scan_init(void) {
    struct bt_le_scan_param scan_param = {
        .type       = BT_HCI_LE_SCAN_PASSIVE,
        .filter_dup = BT_HCI_LE_SCAN_FILTER_DUP_ENABLE,
        .interval   = 0x00A0,
        .window     = 0x00A0,
    };

    struct bt_scan_init_param scan_init = {
        .connect_if_match = false,
        .scan_param       = &scan_param,
        .conn_param       = BT_LE_CONN_PARAM_DEFAULT,
    };

    bt_scan_init(&scan_init);
    bt_scan_cb_register(&scan_cb);
}

static struct bt_conn_cb conn_callbacks = {
    .connected    = connected,
    .disconnected = disconnected,
};

void ble_init(void) {
    int err;

    LOG_INF("Initializing Bluetooth..");
    err = bt_enable(NULL);
    if (err) {
        LOG_INF("Bluetooth init failed (err %d)", err);
        return;
    }

    scan_init();
    bt_conn_cb_register(&conn_callbacks);

    // start BLE scanning
    err = bt_le_scan_start(BT_LE_SCAN_PASSIVE, device_found);
    if (err) {
        LOG_ERR("Scanning failed to start, err %d", err);
    } else {
        LOG_INF("scan started");
    }
}